package example.stateless.hello;

import java.io.*;

import javax.ejb.*;
import javax.naming.*;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * Client servlet querying a server about a greeting.
 *
 * <h3>JNDI client configuration</h3>
 *
 * <code><pre>
 * &lt;jndi-link>
 *   &lt;jndi-name>java:comp/env/ejb&lt;/jndi-name>
 *   &lt;jndi-factory>com.caucho.ejb.BurlapContextFactory&lt;/jndi-factory>
 *   &lt;init-param java.naming.provider.url="http://localhost:8080/cmp/example"/>
 * &lt;/jndi-link>
 * </pre></code>
 */
public class HelloServlet extends HttpServlet {
  /**
   * The servlet stores the home interface after the initial lookup.
   * Since the home interface never changes, caching the lookup will save
   * some performance.
   */
  private HelloHome helloHome;

  /**
   * The init method looks up the HelloHome interface using JNDI and
   * stores it in a servlet variable.
   */
  public void init()
    throws ServletException
  {
    try {
      /*
       * Look up the remote EJB context using JNDI.
       * The context contains the remote EJB beans.
       *
       * Since we're using the remote interface, it uses the ejb.
       */
      Context ejb = (Context) new InitialContext().lookup("java:comp/env/ejb");
      /*
       * Look up the hello home.  "stateless-hello" is the ejb-name
       * in the deployment descriptor.
       */
      helloHome = (HelloHome) ejb.lookup("stateless-hello");
    } catch (Exception e) {
      throw new ServletException(e);
    }
  }

  /**
   * Calls the hello method.  The Hello interface is created using the
   * create method.
   *
   * @param request the servlet request object.
   * @param response the servlet response object.
   */
  public void doGet(HttpServletRequest request,
                    HttpServletResponse response)
    throws IOException, ServletException
  {
    try {
      PrintWriter out = response.getWriter();

      response.setContentType("text/html");

      Hello hello = helloHome.create();

      out.print("Message: ");
      out.println(hello.hello());
    } catch (CreateException e) {
      throw new ServletException(e);
    }
  }
}
